# ruff: noqa: E501 <- line too long
"""Region definitions for the Sheerwater Benchmarking project.

The regions are defined as follows:
- countries
    - 242 unique countries
- subregions:
  - 'central_america', 'caribbean', 'north_america',
  - 'south_eastern_asia', 'western_asia', 'south_asia',
  - 'eastern_asia', 'central_asia', 'seven_seas_open_ocean',
  - 'south_america', 'eastern_africa', 'north_africa',
  - 'middle_africa', 'southern_africa', 'western_africa',
  - 'western_europe', 'eastern_europe', 'north_europe',
  - 'south_europe', 'melanesia', 'australia_and_new_zealand',
  - 'polynesia', 'micronesia', 'antarctica'
- regions_un: United Nations regions
  - 'americas', 'asia', 'africa', 'europe', 'oceania', 'antarctica'
- regions_wb
  - 'latin_america_and_caribbean', 'north_america',
  - 'europe_and_central_asia', 'east_asia_and_pacific', 'south_asia',
  - 'middle_east_and_north_africa', 'sub_saharan_africa', 'antarctica'
- continents
    - 'north_america', 'asia', 'south_america', 'africa', 'europe',
    - 'oceania', 'antarctica', 'seven_seas_open_ocean'
- meteorological_zones
  - 'tropics', 'extratropics'
- hemispheres
  - 'northern_hemisphere', 'southern_hemisphere'
- global:
  - 'global'
"""
import os

import geopandas as gpd
from shapely.geometry import box

from nuthatch import cache

from .general_utils import load_object


def clean_name(name):
    """Clean a name to make matching easier."""
    return name.lower().replace(' ', '_')


# Define the class of standard regions
standard_regions = ['country', 'continent', 'subregion', 'region_un', 'region_wb']

# Additionally, allow the construction of custom regions by country list or lat / lon bounding box
custom_regions = {
    'sheerwater_region': {
        'nimbus_east_africa': {
            'countries': ['kenya', 'burundi', 'rwanda', 'tanzania', 'uganda'],
        },
        'nimbus_west_africa': {
            'countries': ['benin', 'burkina_faso', 'cape_verde', 'ivory_coast', 'the_gambia', 'ghana', 'guinea', 'guinea-bissau', 'liberia', 'mali', 'mauritania', 'niger', 'nigeria', 'senegal', 'sierra_leone', 'togo'],
        },
        'conus': {
            'countries': ['united_states_of_america'],
        }
    },
    'meteorological_zone': {
        'tropics': {
            'lons': [-180.0, 180.0],
            'lats': [-23.5, 23.5]
        },
        'extratropics_northern': {
            'lons': [-180.0, 180.0],
            'lats': [-90.0, -23.5],
        },
        'extratropics_southern': {
            'lons': [-180.0, 180.0],
            'lats': [23.5, 90.0],
        }
    },
    'hemisphere': {
        'northern_hemisphere': {
            'lons': [-180.0, 180.0],
            'lats': [0.0, 90.0]
        },
        'southern_hemisphere': {
            'lons': [-180.0, 180.0],
            'lats': [-90.0, 0.0]
        }
    },
    'global': {
        'global': {
            'lons': [-180.0, 180.0],
            'lats': [-90.0, 90.0]
        }
    }
}


# By enabling caching this can be cached locally
@cache(cache=True, cache_args=[])
def get_country_gdf():
    """Get the country GeoDataFrame."""
    # World geojson downloaded from https://geojson-maps.kyd.au
    filepath = 'gs://sheerwater-public-datalake/regions/world_50m.geojson'
    country_gdf = gpd.read_file(load_object(filepath))
    country_gdf = country_gdf[['name_en', 'continent', 'region_un', 'subregion', 'region_wb', 'geometry']]
    # Clean string columns for consistent, lowercase, and underscore-separated names
    for col in ["name_en", "continent", "region_un", "subregion", "region_wb"]:
        cleaned_col = "country" if col == "name_en" else col
        country_gdf[cleaned_col] = country_gdf[col].apply(clean_name)
    return country_gdf


def get_region_level(region):
    """Get the level of a region and the regions at that level."""
    country_gdf = get_country_gdf(memoize=True)
    if region is None:
        region = 'global'

    # Find which region level the region is in, and get a list of regions at that level
    # If a region could possibly match several region levels, it will return the first
    found = False
    for level in standard_regions:
        if found:
            break
        if region == level:
            region_level = level
            regions = country_gdf[region_level].unique()
            found = True
        if region in country_gdf[level].unique():
            region_level = level
            regions = [region]
            found = True
    for level, data in custom_regions.items():
        if found:
            break
        if region == level:
            region_level = level
            regions = data.keys()
            found = True
        elif region in data.keys():
            region_level = level
            regions = [region]
            found = True
    if not found:
        raise ValueError(f"Region {region} not found")
    return region_level, regions


def region_data(region):
    """Get the boundary shapefile for a given region.

    Args:
        region (str): The region to get the data for. Can be either
            a region level or a specific region within that level. So, for example,
            both 'countries' and 'indonesia' are valid regions.

    Returns:
        gdf (gpd.GeoDataFrame): A GeoDataFrame for the region, with columns:
            - 'region_name': the name of the region,
            - 'region_geometry': its geometry as a shapely object.
    """
    # Standardize input region name
    region = clean_name(region)

    # Get the region data needed to form the GeoDataFrame
    country_gdf = get_country_gdf(memoize=True)
    region_level, regions = get_region_level(region)

    # Form the GeoDataFrame for the region
    region_names = []
    region_geometries = []
    for reg in regions:
        if region_level in standard_regions:
            # Handle standard regions
            region_names.append(reg)
            geometry = country_gdf[country_gdf[region_level] == reg].geometry.union_all()
            region_geometries.append(geometry)
        else:
            # Handle custom regions
            data = custom_regions[region_level][reg]
            if 'lats' in data and 'lons' in data:
                # Create a shapefile (GeoDataFrame) from lat/lon boundaries as a rectangular Polygon
                tol = 0.005
                lons = data['lons']
                lats = data['lats']
                region_box = box(lons[0] - tol, lats[0] - tol, lons[1] + tol, lats[1] + tol)
                region_names.append(reg)
                region_geometries.append(region_box)
            elif 'countries' in data:
                countries = [clean_name(country) for country in data['countries']]
                region_gdf = country_gdf[country_gdf['country'].isin(countries)]
                if len(countries) != len(region_gdf):
                    raise ValueError(f"Some countries were not found: {set(countries) - set(region_gdf['country'])}")
                geometry = region_gdf.geometry.union_all()
                region_names.append(reg)
                region_geometries.append(geometry)
            else:
                raise ValueError(f"Poorly formatted custom region entry: {data}")

    gdf = gpd.GeoDataFrame({'region_name': region_names, 'region_geometry': region_geometries})
    gdf = gdf.set_geometry("region_geometry")
    gdf = gdf.set_crs("EPSG:4326")
    return gdf


def to_name(name):
    """Convert a name to a more readable format."""
    if name == 'mae':
        return 'Mean Absolute Error'
    elif name == 'mse':
        return 'Mean Squared Error'
    elif name == 'rmse':
        return 'Root Mean Squared Error'
    elif name == 'bias':
        return 'Bias'
    elif name == 'crps':
        return 'Continuous Ranked Probability Score'
    elif name == 'brier':
        return 'Brier Score'
    elif name == 'smape':
        return 'Symmetric Mean Absolute Percentage Error'
    elif name == 'mape':
        return 'Mean Absolute Percentage Error'
    elif name == 'seeps':
        return 'Spatial Error in Ensemble Prediction Scale'
    elif 'heidke' in name:
        return f'Heidke Skill Score at {", ".join([mm.split("-")[-1] for mm in name.split("-")[:-1]])} mm'
    elif 'pod' in name:
        return f'Probability of Detection at {name.split("-")[-1]} mm'
    elif 'far' in name:
        return f'False Alarm Rate at {name.split("-")[-1]} mm'
    elif 'ets' in name:
        return f'Equitable Threat Score at {name.split("-")[-1]} mm'
    elif 'csi' in name:
        return f'Critical Success Index at {name.split("-")[-1]} mm'
    elif 'frequency_bias' in name:
        return f'Frequency Bias at {name.split("-")[-1]} mm'
    elif name == 'acc':
        return 'Anomaly Correlation Coefficient'
    elif name == 'pearson':
        return 'Pearson Correlation'
    else:
        return name


def bounds(variable):
    """Get the bounds for a variable."""
    if variable == 'mae':
        return (0.0, None)
    elif variable == 'mse':
        return (0.0, None)
    elif variable == 'rmse':
        return (0.0, None)
    elif variable == 'bias':
        return (None, None)
    elif variable == 'crps':
        return (0.0, None)
    elif variable == 'brier':
        return (0.0, None)
    elif variable == 'smape':
        return (0.0, None)
    elif variable == 'mape':
        return (0.0, None)
    elif variable == 'seeps':
        return (0.0, 3.0)
    elif 'heidke' in variable:
        return (0.0, 1.0)
    elif 'pod' in variable:
        return (0.0, 1.0)
    elif 'far' in variable:
        return (0.0, 1.0)
    elif 'ets' in variable:
        return (0.0, 1.0)
    elif 'csi' in variable:
        return (0.0, 1.0)
    elif 'frequency_bias' in variable:
        return (None, None)
    elif variable == 'acc':
        return (-1.0, 1.0)
    elif variable == 'pearson':
        return (-1.0, 1.0)
    elif variable == 'coverage':
        return (0.0, None)
    else:
        return (None, None)


def plot_by_region(ds, region, variable, file_string='none', title='Regional Map'):
    """Plot a variable from an xarray dataset by region.

    Args:
        ds: xarray DataArray or Dataset with a 'region' coordinate
        region: Region level or specific region name to plot
        variable: Variable name if ds is a Dataset (optional)
        file_string: File path string for saving the plot
        title: Title for the plot

    Returns:
        matplotlib axes object
    """
    import matplotlib.pyplot as plt
    import numpy as np

    # Get the region GeoDataFrame and metric bounds
    gdf = region_data(region)
    # Extract the data values
    try:
        data = ds[variable]
    except KeyError:
        data = ds

    # Convert to numpy array if it's a dask array
    if hasattr(data, 'compute'):
        data = data.compute()

    # Extract values for each region in the GeoDataFrame
    values = []
    for region_name in gdf.region_name:
        try:
            # Select the region and extract the scalar value
            region_value = data.sel(region=region_name)
            # Handle case where selection might still have dimensions
            if region_value.size > 1:
                # If there are multiple values, take the first or mean
                region_value = float(region_value.values.flatten()[0])
            else:
                region_value = float(region_value.values)
            values.append(region_value)
        except (KeyError, ValueError):
            # Region not found in dataset, use NaN
            values.append(np.nan)

    # Add values to GeoDataFrame
    gdf['value'] = values

    # Calculate 95th percentile for vmax (excluding NaNs)
    vmin, vmax = bounds(variable)
    values_array = np.array(values)
    valid_values = values_array[~np.isnan(values_array)]
    if vmin is None:
        vmin = np.percentile(valid_values, 5)
    if vmax is None:
        vmax = np.percentile(valid_values, 95)

    # Reproject to a better projection for world maps (Robinson projection)
    # Robinson is good for world maps, preserves area relationships well
    gdf_projected = gdf.to_crs('ESRI:54030')  # Robinson projection

    # Create the plot
    fig, ax = plt.subplots(1, 1, figsize=(14, 8))

    # Plot the choropleth map (without legend, we'll add it manually)
    gdf_projected.plot(column='value', ax=ax, legend=False,
                       cmap='viridis', edgecolor='black', linewidth=0.5,
                       vmin=vmin,
                       vmax=vmax,
                       missing_kwds={'color': 'red', 'edgecolor': 'black'})

    # Create colorbar with same height as plot and larger tick labels
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.1)
    sm = plt.cm.ScalarMappable(cmap=plt.cm.viridis,
                               norm=plt.Normalize(vmin=vmin, vmax=vmax))
    sm.set_array([])
    cbar = plt.colorbar(sm, cax=cax)
    cbar.ax.tick_params(labelsize=12)

    # Remove axis labels (not meaningful in projected coordinates)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    # Set title
    ax.set_title(title, fontsize=14, pad=10)

    plt.tight_layout()

    # Ensure colorbar height matches plot height exactly
    ax_pos = ax.get_position()
    cax_pos = cax.get_position()
    cax.set_position([cax_pos.x0, ax_pos.y0, cax_pos.width, ax_pos.height])
    # Save
    if not os.path.exists('metric_maps'):
        os.makedirs('metric_maps')
    plt.savefig(f'metric_maps/{variable}_{region}_{file_string}_map.png')
    return ax



"""Tasks models for the Sheerwater benchmarking project."""
from .spw import spw_precip_preprocess, spw_rainy_onset

# Use __all__ to define what is part of the public API.
__all__ = ["spw_rainy_onset", "spw_precip_preprocess"]



"""Interface for graphcast forecasts."""
import numpy as np
import pandas as pd
import xarray as xr
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.utils import dask_remote, lon_base_change, regrid, roll_and_agg, shift_by_days
from sheerwater.interfaces import forecast as sheerwater_forecast, spatial


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'init_hour', 'grid'],
       backend_kwargs={'chunking': {"lat": 721, "lon": 1440, "lead_time": 1, "time": 30}})
def graphcast_daily(start_time, end_time, variable, init_hour=0, grid='global0_25', mask=None, region='global'):  # noqa: ARG001
    """A daily Graphcast forecast."""
    if init_hour != 0:
        raise ValueError("Only 0 init hour supported.")

    # Read the three years for gcloud
    ds1 = xr.open_zarr(
        'gs://weathernext/59572747_4_0/zarr/99140631_1_2020_to_2021/forecasts_10d/date_range_2019-12-01_2021-01-22_6_hours.zarr',
        chunks='auto',
        decode_timedelta=True)
    ds1 = ds1.rename({'prediction_timedelta': 'lead_time',
                     '2m_temperature': 'tmp2m', 'total_precipitation_6hr': 'precip'})
    ds1 = ds1[[variable]]
    ds1 = ds1.sel({'time': slice(pd.to_datetime("2019-12-01"), pd.to_datetime("2020-11-30"))})

    ds2 = xr.open_zarr(
        'gs://weathernext/59572747_4_0/zarr/99140631_2_2021_to_2022/forecasts_10d/date_range_2020-12-01_2022-01-22_6_hours.zarr',
        decode_timedelta=True,
        chunks='auto')
    ds2 = ds2.rename({'prediction_timedelta': 'lead_time',
                     '2m_temperature': 'tmp2m', 'total_precipitation_6hr': 'precip'})
    ds2 = ds2[[variable]]
    ds2 = ds2.sel({'time': slice(pd.to_datetime("2020-12-01"), pd.to_datetime("2021-12-31"))})

    ds3 = xr.open_zarr(
        'gs://weathernext/59572747_4_0/zarr/132880704_2022_to_2023/1/forecasts_10d/date_range_2022-01-01_2023-01-01_12_hours.zarr',
        decode_timedelta=True,
        chunks='auto')
    ds3 = ds3.rename({'prediction_timedelta': 'lead_time',
                     '2m_temperature': 'tmp2m', 'total_precipitation_6hr': 'precip'})
    ds3 = ds3[[variable]]
    ds3 = ds3.sel({'time': slice(pd.to_datetime("2022-01-01"), pd.to_datetime("2023-01-01"))})

    # concat them together
    ds = xr.concat([ds1, ds2, ds3], dim='time')

    # Select only the midnight initialization times
    ds = ds.where(ds.time.dt.hour == init_hour, drop=True)

    # Convert units
    K_const = 273.15
    if variable == 'tmp2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(lead_time='1D').mean(dim='lead_time')
    elif variable == 'precip':
        ds[variable] = ds[variable] * 1000.0
        ds.attrs.update(units='mm')
        # Convert from 6hrly to daily precip, robust to different numbers of 6hrly samples in a day
        ds = ds.resample(lead_time='1D').mean(dim='lead_time') * 4.0
        # Can't have precip less than zero (there are some very small negative values)
        ds = np.maximum(ds, 0)
    else:
        raise ValueError(f"Variable {variable} not implemented.")

    # Shift the lead back 6 hours + init time to align with the start date convention
    ds['lead_time'] = ds['lead_time'] - np.timedelta64(6+init_hour, 'h')

    # Convert lat/lon
    ds = lon_base_change(ds)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'init_hour', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 1, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'lead_time': 1, 'time': 30}
               },
           }
})
def graphcast_daily_wb(start_time, end_time, variable, init_hour=0, grid='global0_25', mask=None, region='global'):  # noqa: ARG001
    """A daily Graphcast forecast."""
    if init_hour != 0:
        raise ValueError("Only 0 init hour supported.")

    # Read the three years for gcloud
    if grid == 'global0_25':
        filename1 = 'gs://weatherbench2/datasets/graphcast/2018/date_range_2017-11-16_2019-02-01_12_hours.zarr'
        filename2 = 'gs://weatherbench2/datasets/graphcast/2020/date_range_2019-11-16_2021-02-01_12_hours.zarr'
    elif grid == 'global1_5':
        # Global 1.5 degree grid is fractionally off from the standard 1.5 grid
        filename1 = 'gs://weatherbench2/datasets/graphcast/2018/date_range_2017-11-16_2019-02-01_12_hours-240x121_equiangular_with_poles_conservative.zarr'
        filename2 = 'gs://weatherbench2/datasets/graphcast/2020/date_range_2019-11-16_2021-02-01_12_hours-240x121_equiangular_with_poles_conservative.zarr'
    else:
        raise ValueError(f"Grid {grid} not implemented.")

    ds1 = xr.open_zarr(filename1,
                       chunks='auto',
                       decode_timedelta=True)
    ds1 = ds1.rename({'prediction_timedelta': 'lead_time',
                     '2m_temperature': 'tmp2m', 'total_precipitation_6hr': 'precip'})
    if 'latitude' in ds1.dims:
        ds1 = ds1.rename({'latitude': 'lat', 'longitude': 'lon'})
    ds1 = ds1[[variable]]

    ds2 = xr.open_zarr(filename2,
                       decode_timedelta=True,
                       chunks='auto')
    ds2 = ds2.rename({'prediction_timedelta': 'lead_time',
                     '2m_temperature': 'tmp2m', 'total_precipitation_6hr': 'precip'})
    if 'latitude' in ds2.dims:
        ds2 = ds2.rename({'latitude': 'lat', 'longitude': 'lon'})
    ds2 = ds2[[variable]]

    # concat them together
    ds = xr.concat([ds1, ds2], dim='time')
    ds = ds.sortby('time')

    # Select only the midnight initialization times
    ds = ds.where(ds.time.dt.hour == init_hour, drop=True)

    # Round the lats to two decimal places
    ds['lat'] = np.round(ds.lat, decimals=2)
    ds['lon'] = np.round(ds.lon, decimals=2)

    # Convert units
    K_const = 273.15
    if variable == 'tmp2m':
        ds[variable] = ds[variable] - K_const
        ds.attrs.update(units='C')
        ds = ds.resample(lead_time='1D').mean(dim='lead_time')
    elif variable == 'precip':
        ds[variable] = ds[variable] * 1000.0
        ds.attrs.update(units='mm')
        # Convert from 6hrly to daily precip, robust to different numbers of 6hrly samples in a day
        ds = ds.resample(lead_time='1D').mean(dim='lead_time') * 4.0
        # Can't have precip less than zero (there are some very small negative values)
        ds = np.maximum(ds, 0)
    else:
        raise ValueError(f"Variable {variable} not implemented.")

    # Shift the lead back 6 hours + init time to align with the start date convention
    ds['lead_time'] = ds['lead_time'] - np.timedelta64(6+init_hour, 'h')

    # Convert lat/lon
    ds = lon_base_change(ds)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'init_hour', 'grid'],
       cache_disable_if={'grid': 'global0_25'},
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 1, "time": 1000},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'lead_time': 1, 'time': 30}
               },
           }
})
def graphcast_daily_regrid(start_time, end_time, variable, init_hour=0,
                           grid='global0_25', mask=None,
                           region='global'):  # noqa: ARG001
    """Regrid for the original Weathernext datasource."""
    ds = graphcast_daily(start_time, end_time, variable, init_hour=init_hour,
                         grid='global0_25', mask=mask, region=region)
    if grid == 'global0_25':
        return ds

    # Regrid onto appropriate grid
    ds = regrid(ds, grid, base='base180', method='conservative', output_chunks={"lat": 121, "lon": 240}, region=region)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['variable', 'agg_days', 'grid'],
       cache_disable_if={'agg_days': 1},
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 10, "time": 100},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'lead_time': 1, 'time': 30}
               },
           }
})
def graphcast_wb_rolled(start_time, end_time, variable, agg_days, grid='global0_25', mask=None, region='global'):
    """A rolled and aggregated Graphcast forecast."""
    # Grab the init 0 forecast; don't need to regrid
    ds = graphcast_daily_wb(start_time, end_time, variable, init_hour=0, grid=grid, mask=mask, region=region)
    ds = roll_and_agg(ds, agg=agg_days, agg_col="lead_time", agg_fn="mean")
    return ds


@dask_remote
@sheerwater_forecast()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'prob_type', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365, 'lead_time': 1, 'member': 1}})
def graphcast(start_time=None, end_time=None, variable="precip", agg_days=1, prob_type='deterministic',
              grid='global1_5', mask='lsm', region="global"):  # noqa: ARG001
    """Final Graphcast interface."""
    if prob_type != 'deterministic':
        raise NotImplementedError("Only deterministic forecast implemented for graphcast")

    # Get the data with the right days
    forecast_start = shift_by_days(start_time, -15) if start_time is not None else None
    forecast_end = shift_by_days(end_time, 15) if end_time is not None else None

    # Get the data with the right days
    ds = graphcast_wb_rolled(forecast_start, forecast_end, variable,
                             agg_days=agg_days, grid=grid, mask=mask,
                             region=region)
    ds = ds.assign_attrs(prob_type="deterministic")

    # Rename to standard naming
    ds = ds.rename({'time': 'init_time', 'lead_time': 'prediction_timedelta'})

    return ds



"""Utility functions for benchmarking."""
from .data_utils import get_anomalies, regrid, roll_and_agg
from .forecaster_utils import convert_init_time_to_pred_time, get_variable
from .general_utils import load_netcdf, load_object, load_zarr, plot_ds, plot_ds_map, run_in_parallel, write_zarr
from .grouping_utils import groupby_region, groupby_time, latitude_weights, mean_or_sum
from .region_utils import region_data, get_region_level, plot_by_region
from .remote import dask_remote, start_remote
from .secrets import cdsapi_secret, ecmwf_secret, gap_secret, salient_secret, tahmo_secret, huggingface_read_token
from .space_utils import (
    apply_mask,
    base180_to_base360,
    base360_to_base180,
    check_bases,
    clip_region,
    get_globe_slice,
    get_grid,
    get_grid_ds,
    is_wrapped,
    lon_base_change,
    snap_point_to_grid,
)
from .task_utils import first_satisfied_date
from .time_utils import (
    add_dayofyear,
    assign_grouping_coordinates,
    convert_group_to_time,
    date_mean,
    doy_mean,
    generate_dates_in_between,
    get_dates,
    is_valid_forecast_date,
    pad_with_leapdays,
    shift_by_days,
)

# Use __all__ to define what is part of the public API.
__all__ = [
    "dask_remote",
    "start_remote",
    "cdsapi_secret",
    "ecmwf_secret",
    "salient_secret",
    "tahmo_secret",
    "gap_secret",
    "roll_and_agg",
    "get_anomalies",
    "regrid",
    "load_netcdf",
    "write_zarr",
    "load_zarr",
    "load_object",
    "plot_ds",
    "plot_ds_map",
    "run_in_parallel",
    "get_region_level",
    "region_data",
    "get_grid",
    "get_grid_ds",
    "is_wrapped",
    "base360_to_base180",
    "base180_to_base360",
    "check_bases",
    "clip_region",
    "lon_base_change",
    "get_globe_slice",
    "apply_mask",
    "is_valid_forecast_date",
    "generate_dates_in_between",
    "get_dates",
    "pad_with_leapdays",
    "add_dayofyear",
    "shift_by_days",
    "groupby_region",
    "latitude_weights",
    "mean_or_sum",
    "groupby_time",
    "assign_grouping_coordinates",
    "convert_group_to_time",
    "date_mean",
    "doy_mean",
    "get_variable",
    "convert_init_time_to_pred_time",
    "first_satisfied_date",
    "snap_point_to_grid",
    "plot_by_region",
    "huggingface_read_token"
]



"""Functions to fetch and process data from the ECMWF IRI data library.

NOTE: these are legacy functions and no longer actively maintained.
"""
import os
import ssl
import sys
import time
from datetime import datetime

import dask
import dateparser
import pandas as pd
import requests
import xarray as xr
from nuthatch import cache
from nuthatch.processors import timeseries
from urllib3 import poolmanager

from sheerwater.utils import (
    dask_remote,
    ecmwf_secret,
    get_dates,
    get_grid,
    is_valid_forecast_date,
    lon_base_change,
    regrid,
    roll_and_agg,
)
from sheerwater.interfaces import spatial


########################################################################
# IRI download utility functions
########################################################################
class TLSAdapter(requests.adapters.HTTPAdapter):
    """Transport adapter that allows us to use TLSv1.2."""

    def init_poolmanager(self, connections, maxsize, block=False):
        """Create and initialize the urllib3 PoolManager."""
        ctx = ssl.create_default_context()
        ctx.set_ciphers('DEFAULT@SECLEVEL=1')
        self.poolmanager = poolmanager.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_version=ssl.PROTOCOL_TLS,
            ssl_context=ctx)


def download_url(url, timeout=600, retry=3, cookies={}):
    """Download URL, waiting some time between retries."""
    r = None
    session = requests.session()
    session.mount('https://', TLSAdapter())

    for i in range(retry):
        try:
            r = session.get(url, timeout=timeout, cookies=cookies)
            return r
        except requests.exceptions.Timeout as e:
            # Wait until making another request
            if i == retry - 1:
                raise e
            print(f"Request to url {url} has timed out. Trying again...")
            time.sleep(3)
    print(f"Failed to retrieve file after {retry} attempts. Stopping...")


########################################################################
#  IRI ECMWF download and process functions
########################################################################
@dask_remote
@cache(cache_args=['time', 'variable', 'forecast_type', 'run_type', 'grid'],
       backend_kwargs={'chunking': {'lat': 121, 'lon': 240, 'lead_time': 46, 'model_run': 1,
                                    'start_date': 969, 'model_issuance_date': 1}})
def single_iri_ecmwf(time, variable, forecast_type,
                     run_type="average", grid="global1_5",
                     verbose=True):
    """Fetches forecast data from the IRI ECMWF dataset.

    Args:
        time (str): The date to fetch data for (by day).
        variable (str): The weather variable to fetch.
        forecast_type (str): The type of forecast to fetch. One of "forecast" or "reforecast".
        run_type (str): The type of run to fetch. One of:
            - average: to download the averaged of the perturbed runs
            - control: to download the control forecast
            - perturbed: to download all perturbed runs
            - [int 0-50]: to download a specific  perturbed run
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
        verbose (bool): Whether to print verbose output.
    """
    if variable == "tmp2m":
        weather_variable_name_on_server = "2m_above_ground/.2t"
    elif variable == "precip":
        weather_variable_name_on_server = "sfc_precip/.tp"
    else:
        raise ValueError("Invalid weather variable.")

    forecast_runs = "control" if run_type == "control" else "perturbed"
    leads_id = "LA" if variable == "tmp2m" else "L"
    average_model_runs_url = "[M]average/" if run_type == "average" else ""
    single_model_run_url = f"M/({run_type})VALUES/" if isinstance(run_type, int) else ""

    lons, lats, grid_size = get_grid(grid)
    restrict_lat_url = f"Y/{lats[0]}/{grid_size}/{lats[-1]}/GRID/"
    restrict_lon_url = f"X/{lons[0]}/{grid_size}/{lons[-1]}/GRID/"

    if variable == "tmp2m":
        # Convert from Kelvin to Celsius.
        differences_url = ""
        convert_units_url = "(Celsius_scale)/unitconvert/"
    else:
        differences_url = "[L]differences/"
        convert_units_url = "c://name//water_density/def/998/(kg/m3)/:c/div/(mm)/unitconvert/"

    if verbose:
        print("ecmwf")

    date = dateparser.parse(time)
    day, month, year = datetime.strftime(date, "%d,%b,%Y").split(",")

    restrict_forecast_date_url = f"S/({day} {month} {year})VALUES/"

    if not is_valid_forecast_date("ecmwf", forecast_type, date):
        if verbose:
            print(
                f"Skipping: {day} {month} {year} (not a valid {forecast_type} date)."
            )
        return None

    URL = (
        f"https://iridl.ldeo.columbia.edu/SOURCES/.ECMWF/.S2S/.ECMF/"
        f".{forecast_type}/.{forecast_runs}/.{weather_variable_name_on_server}/"
        f"{average_model_runs_url}"
        f"{single_model_run_url}"
        f"{restrict_forecast_date_url}"
        f"{restrict_lat_url}"
        f"{restrict_lon_url}"
        f"{convert_units_url}"
        f"{differences_url}"
        f"data.nc"
    )

    os.makedirs('./temp', exist_ok=True)
    file = f"./temp/{variable}-{grid}-{run_type}-{forecast_type}-{time}.nc"
    r = download_url(URL, cookies={"__dlauth_id": ecmwf_secret()})
    if r.status_code == 200 and r.headers["Content-Type"] == "application/x-netcdf":
        if verbose:
            print(f"Downloading: {day} {month} {year}.")
        with open(file, "wb") as f:
            f.write(r.content)

        if verbose:
            print(f"-done (downloaded {sys.getsizeof(r.content) / 1024:.2f} KB).\n")
    elif r.status_code == 404:
        print(f"Data for {day} {month} {year} is not available for model ecmwf.\n")
        return None
    else:
        raise ValueError(f"Failed to download data for {day} {month} {year} for model ecmwf.")

    rename_dict = {
        "S": "start_date",
        f"{leads_id}": "lead_time",
        "X": "lon",
        "Y": "lat",
    }

    if variable == "precip":
        rename_dict["ratio"] = "precip"
    else:
        rename_dict["2t"] = "tmp2m"

    try:
        # Read the data and return individual datasets
        if forecast_type == "forecast":
            ds = xr.open_dataset(file, engine="netcdf4")
        else:
            ds = xr.open_dataset(file, decode_times=False, engine="netcdf4")

            # Manually decode the time variable
            ds['S'] = pd.to_datetime(
                ds['S'].values, unit="D", origin=pd.Timestamp("1960-01-01"))

            model_issuance_day = ds['S.day'].values[0]
            model_issuance_month = ds['S.month'].values[0]
            model_issuance_date_in_1960 = pd.Timestamp(
                f"1960-{model_issuance_month}-{model_issuance_day}")

            # While hdates refer to years (for which the model issued in ds["S"] is initialized),
            # its values are given as months until the middle of the year, so 6 months are subtracted
            # to yield the beginning of the year.
            ds['hdate'] = pd.to_datetime(
                [model_issuance_date_in_1960 + pd.DateOffset(months=x-6) for x in ds['hdate'].values])

            # Drop future forecast dates, which are all NaNs
            ds = ds.dropna(dim="hdate", how="all")
            if ds.sizes["hdate"] == 0:
                # If no forecast dates are available, return None
                print(f"No data found for: {day} {month} {year} (a valid {forecast_type} date).")
                return None

            # Reforecast-specific renaming
            rename_dict["hdate"] = "start_date"
            rename_dict["S"] = "model_issuance_date"
    except OSError:
        print(f"Failed to load data for: {day} {month} {year} (a valid {forecast_type} date).")
        return None

    # Deal with model runs
    if "M" in ds and ds.sizes["M"] == 1:
        ds = ds.squeeze("M")
    elif "M" in ds:
        rename_dict["M"] = "model_run"

    # Rename columns to standard names
    ds = ds.rename(rename_dict)

    ds = ds.compute()
    os.remove(file)
    return ds


@dask_remote
@cache(cache_args=['time', 'variable', 'forecast_type', 'run_type', 'grid'],
       backend_kwargs={'chunking': {'lat': 121, 'lon': 240, 'lead_time': 46, 'model_run': 1,
                                    'start_year': 20, 'model_issuance_date': 1}})
def single_iri_ecmwf_dense(time, variable, forecast_type,
                           run_type="average", grid="global1_5",
                           verbose=True):
    """Fetches a single IRI ECMWF forecast and then converts to dense format.

    Converts the start_date to start date year. This allows a dense array post merging because
    the data within the year is always the same as the model issuance date.
    eliminating the sparsity of start_dates without model issuance dates.

    Interface is the same as single_iri_ecmwf.
    """
    ds = single_iri_ecmwf(time, variable, forecast_type, run_type, grid, verbose)

    if ds is None:
        return None

    # Convert start_date to start_year
    ds = ds.assign_coords(start_date=ds['start_date.year'])
    ds = ds.rename({'start_date': 'start_year'})

    return ds


@dask_remote
@timeseries(timeseries=['start_date', 'model_issuance_date'])
@cache(cache=False,
       cache_args=['variable', 'forecast_type', 'run_type', 'grid'],
       backend_kwargs={'chunking': {'lat': 121, 'lon': 240, 'lead_time': 46, 'start_date': 969,
                                    'model_run': 1, 'start_year': 29, 'model_issuance_date': 1}})
def iri_ecmwf(start_time, end_time, variable, forecast_type,
              run_type="average", grid="global1_5", verbose=False):
    """Fetches forecast data from the ECMWF IRI dataset.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        forecast_type (str): The type of forecast to fetch. One of "forecast" or "reforecast".
        run_type (str): The type of run to fetch. One of:
            - average: to download the averaged of the perturbed runs
            - control: to download the control forecast
            - perturbed: to download all perturbed runs
            - [int 0-50]: to download a specific  perturbed run
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
        verbose (bool): Whether to print verbose output.
    """
    # Read and combine all the data into an array
    target_dates = get_dates(start_time, end_time,
                             stride="day", return_string=True)

    # Get correct single function
    fn = single_iri_ecmwf if forecast_type == "forecast" else single_iri_ecmwf_dense
    datasets = []
    for date in target_dates:
        ds = dask.delayed(fn)(
            date, variable, forecast_type, run_type, grid, verbose, filepath_only=True)
        datasets.append(ds)
    datasets = dask.compute(*datasets)
    data = [d for d in datasets if d is not None]
    if len(data) == 0:
        return None

    if forecast_type == "forecast":
        x = xr.open_mfdataset(data,
                              engine='zarr',
                              combine="by_coords",
                              parallel=True,
                              chunks={'lat': 121, 'lon': 240, 'lead_time': 46, 'start_date': 969})
        return x
    elif forecast_type == "reforecast":
        x = xr.open_mfdataset(data,
                              engine='zarr',
                              concat_dim='model_issuance_date',
                              combine="nested",
                              parallel=True,
                              chunks={'lat': 121, 'lon': 240, 'lead_time': 46, 'model_run': 1,
                                      'start_year': 20, 'model_issuance_date': 1})

        return x


@dask_remote
@timeseries(timeseries=['start_date', 'model_issuance_date'])
@cache(cache_args=['variable', 'forecast_type', 'grid'],
       backend_kwargs={'chunking': {"lat": 121, "lon": 240, "lead_time": 46,
                                    "start_date": 29, "start_year": 29,
                                    "model_issuance_date": 1}})
def ecmwf_averaged_iri(start_time, end_time, variable, forecast_type, grid="global1_5"):
    """Fetches forecast data from the ECMWF IRI dataset.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        forecast_type (str): The type of forecast to fetch. One of "forecast" or "hindcast".
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
    """
    if grid != 'global1_5':
        raise ValueError("Only the global1_5 grid is supported for ecmwf averaged.")

    # Ensure appropriate chunking for merge
    chunk_dict = {'lat': 121, 'lon': 240, 'lead_time': 46}
    if forecast_type == "reforecast":
        chunk_dict['start_year'] = 29
        chunk_dict['model_issuance_date'] = 1
    else:
        chunk_dict['start_date'] = 29

    # Read and combine all the data into an array
    df_control = iri_ecmwf(start_time, end_time, variable,
                           forecast_type, run_type="control",
                           grid=grid) \
        .rename({f"{variable}": f"{variable}_control"})
    # A note: this rechunking shouldn't be necessary, but it is
    df_control = df_control.chunk(chunk_dict)

    df_pert = iri_ecmwf(start_time, end_time, variable,
                        forecast_type, run_type="average",
                        grid=grid) \
        .rename({f"{variable}": f"{variable}_pert"})
    # A note: this rechunking shouldn't be necessary, but it is
    df_pert = df_pert.chunk(chunk_dict)

    # Combine control and perturbed runs
    if forecast_type == "forecast":
        M = 50.0  # forecast average is made up of 50 ensemble members
    else:
        M = 10.0  # reforecast average is made up of 10 ensemble members

    # Need to run a merge here, because sometimes pert and control
    # are available for different dates
    df = xr.merge([df_control, df_pert], join="outer")

    # Take a weighted average of the control and perturbed runs
    df[variable] = df[f"{variable}_control"] * 1./(M+1.) + df[f"{variable}_pert"] * M/(M+1.)
    df = df.drop([f"{variable}_control", f"{variable}_pert"])

    return df


@dask_remote
@timeseries(timeseries=['start_date', 'model_issuance_date'])
@spatial()
@cache(cache_args=['variable', 'forecast_type', 'grid'],
       cache_disable_if={'grid': 'global1_5'},
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 46,
                        "start_date": 30,
                        "model_issuance_date": 30, "start_year": 1},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'model_issuance_date': 1}
               },
           }
       })
def ecmwf_averaged_regrid(start_time, end_time, variable, forecast_type,
                          grid='global1_5', mask=None,  # noqa: ARG001
                          region='global'):
    """IRI ECMWF average forecast with regridding."""
    ds = ecmwf_averaged_iri(start_time, end_time, variable, forecast_type, grid='global1_5')
    # Convert to base180 longitude
    ds = lon_base_change(ds, to_base="base180")

    if grid == 'global1_5':
        return ds
    # Regrid onto appropriate grid
    ds = regrid(ds, grid, base='base180', grid_chunks={"lat": 120, "lon": 120}, region=region)
    return ds


@dask_remote
@timeseries(timeseries=['start_date', 'model_issuance_date'])
@spatial()
@cache(cache_args=['variable', 'forecast_type', 'agg', 'grid'],
       backend_kwargs={
           'chunking': {"lat": 121, "lon": 240, "lead_time": 46,
                        "start_date": 30,
                        "model_issuance_date": 30, "start_year": 1},
           'chunk_by_arg': {
               'grid': {
                   'global0_25': {"lat": 721, "lon": 1440, 'model_issuance_date': 1}
               },
           }
       })
def ecmwf_rolled(start_time, end_time, variable, forecast_type,
                 agg_days=14, grid="global1_5", mask=None, region='global'):
    """Fetches forecast data from the ECMWF IRI dataset.

    Args:
        start_time (str): The start date to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        forecast_type (str): The type of forecast to fetch.
        agg_days (int): The aggregation period to use, in days.
        grid (str): The grid resolution to fetch the data at.
        mask: Spatial mask to apply.
        region: Region to fetch data for.
        end_time (str): The end date to fetch.
        variable (str): The weather variable to fetch.
        forecast_type (str): The type of forecast to fetch. One of "forecast" or "hindcast".
        agg_days (int): The aggregation period, in days.
        grid (str): The grid resolution to fetch the data at. One of:
            - global1_5: 1.5 degree global grid
        region (str): The region to clip the data to.
    """
    # Read and combine all the data into an array
    ds = ecmwf_averaged_regrid(start_time, end_time, variable, forecast_type, grid=grid, mask=mask, region=region)
    ds = ds.chunk({'lat': 721, 'lon': 1440})
    if forecast_type == "reforecast":
        ds = ds.chunk({'start_year': 29, 'model_issuance_date': 1})
    else:
        ds = ds.chunk({'start_date': 29})

    # Roll and aggregate the data
    ds = roll_and_agg(ds, agg=agg_days, agg_col="lead_time", agg_fn="mean")

    return ds



"""Imerg data product."""
import gcsfs
import xarray as xr
from dateutil import parser
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.utils import dask_remote, regrid, roll_and_agg

from sheerwater.interfaces import data as sheerwater_data, spatial

@dask_remote
@cache(cache_args=['year', 'version'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg_raw(year, version='final'):
    """Concatenated IMERG netcdf files by year.

    We manually download the IMERG netcdfs by going to this website
    https://disc.gsfc.nasa.gov/datasets/GPM_3IMERGDL_07/summary?keywords=%22IMERG%20late%22
    logging in, downloading the filelist, then following the curl fetch instructions
    here https://disc.gsfc.nasa.gov/information/howto?title=How%20to%20Access%20GES%20DISC%20Data%20Using%20wget%20and%20curl,
    then finally uploading them to our datalake with tools/copy_imerg.py.
    """
    # Open the datastore
    fs = gcsfs.GCSFileSystem(project='sheerwater', token='google_default')

    if version == 'final':
        gsf = [fs.open(x) for x in fs.glob(f'gs://sheerwater-datalake/imerg/{year}*.nc')]
    elif version == 'late':
        gsf = [fs.open(x) for x in fs.glob(f'gs://sheerwater-datalake/imerg_late/{year}*.nc')]

    ds = xr.open_mfdataset(gsf, engine='h5netcdf', parallel=True)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['grid', 'version'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg_gridded(start_time, end_time, grid, version, mask=None,  # noqa: ARG001
                  region='global'):
    """Regridded version of whole imerg dataset."""
    years = range(parser.parse(start_time).year, parser.parse(end_time).year + 1)

    datasets = []
    for year in years:
        ds = imerg_raw(year, version, filepath_only=True)
        datasets.append(ds)

    ds = xr.open_mfdataset(datasets,
                           engine='zarr',
                           parallel=True,
                           chunks={'lat': 300, 'lon': 300, 'time': 365})

    ds = ds['precipitation'].to_dataset()
    ds = ds.rename({'precipitation': 'precip'})

    # Regrid if not on the native grid
    if grid != 'imerg' and grid != 'global0_1':
        ds = regrid(ds, grid, base='base180', method='conservative', region=region)

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['grid', 'agg_days', 'version'],
       cache_disable_if={'agg_days': 1},
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg_rolled(start_time, end_time, agg_days, grid, version, mask=None, region='global'):
    """IMERG rolled and aggregated."""
    ds = imerg_gridded(start_time, end_time, grid, version, mask=mask, region=region)
    ds = roll_and_agg(ds, agg=agg_days, agg_col="time", agg_fn='mean')
    return ds


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg_final(start_time=None, end_time=None, variable='precip', agg_days=1,
                grid='global0_25', mask='lsm', region='global'):
    """IMERG Final."""
    if variable not in ['precip']:
        raise NotImplementedError("Only precip and derived variables provided by IMERG.")
    return imerg_rolled(start_time, end_time, agg_days=agg_days, grid=grid, version='final', mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg_late(start_time=None, end_time=None, variable='precip', agg_days=1,
               grid='global0_25', mask='lsm', region='global'):
    """IMERG late."""
    if variable not in ['precip']:
        raise NotImplementedError("Only precip and derived variables provided by IMERG.")
    return imerg_rolled(start_time, end_time, agg_days=agg_days, grid=grid, version='late', mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def imerg(start_time=None, end_time=None, variable='precip', agg_days=1,
          grid='global0_25', mask='lsm', region='global'):
    """Alias for IMERG final."""
    if variable not in ['precip']:
        raise NotImplementedError("Only precip and derived variables provided by IMERG.")
    return imerg_rolled(start_time, end_time, agg_days=agg_days, grid=grid, version='final', mask=mask, region=region)



"""Test the metrics library functions."""
from sheerwater.metrics_library import metric_factory
from sheerwater.utils import start_remote

start_remote(remote_config='large_cluster')


def test_metric_factory():
    """Test the metric factory function."""
    cache_kwargs = {
        'start_time': '2016-01-01',
        'end_time': '2022-12-31',
        'variable': 'precip',
        'agg_days': 7,
        'forecast': 'ecmwf_ifs_er_debiased',
        'truth': 'era5',
        'time_grouping': 'month',
        'spatial': False,
        'grid': 'global1_5',
        'mask': 'lsm',
        'region': 'country',
    }

    met = metric_factory('heidke-1-5-10-20', **cache_kwargs)
    test = met.compute()
    assert test is not None


if __name__ == "__main__":
    start_remote(remote_config='large_cluster', remote_name='test_metrics')
    test_metric_factory()



"""Performance tests for grouped metrics with different space and time groupings."""
import time

import pytest

from sheerwater.metrics import metric
from sheerwater.utils import plot_by_region, start_remote


@pytest.fixture(scope="module")
def setup_remote():
    """Set up remote cluster for tests."""
    start_remote(remote_config='large_cluster')


@pytest.mark.parametrize("region", ["global", "country", "continent", "subregion"])
@pytest.mark.parametrize("time_grouping", [None, "month_of_year", "year", "quarter_of_year"])
def test_grouped_metrics_performance(setup_remote, region, time_grouping):  # noqa: ARG001
    """Test performance of grouped metrics with different space and time groupings."""
    start_time = "2016-01-01"
    end_time = "2022-12-31"
    variable = "precip"
    agg_days = 7
    forecast = "ecmwf_ifs_er_debiased"
    truth = "era5"
    metric_name = "heidke-1-5-10-20"
    grid = "global0_25"
    mask = "lsm"

    start = time.time()
    result = metric(
        start_time=start_time,
        end_time=end_time,
        variable=variable,
        agg_days=agg_days,
        forecast=forecast,
        truth=truth,
        metric_name=metric_name,
        time_grouping=time_grouping,
        spatial=False,
        grid=grid,
        mask=mask,
        region=region,
        recompute=False
    )
    elapsed = time.time() - start

    assert result is not None, f"Metric computation failed for region={region}, time_grouping={time_grouping}"

    print(f"Region: {region}, Time grouping: {time_grouping}, Time: {elapsed:.2f}s")


def test_grouped_metrics_performance_summary():
    """Run a comprehensive performance test and print summary."""
    start_time = "2016-01-01"
    end_time = "2022-12-31"
    variable = "precip"
    agg_days = 7
    forecast = "imerg"
    truth = "ghcn"
    metric_name = "mae"
    grid = "global0_25"
    mask = "lsm"

    regions = ["global", "global", "country", "continent", "subregion"]
    time_groupings = [None, "month_of_year", "year", "month"]

    results = []
    for region in regions:
        for time_grouping in time_groupings:
            start = time.time()
            result = metric(
                start_time=start_time,
                end_time=end_time,
                variable=variable,
                agg_days=agg_days,
                forecast=forecast,
                truth=truth,
                metric_name=metric_name,
                time_grouping=time_grouping,
                spatial=False,
                grid=grid,
                mask=mask,
                region=region,
                recompute=True,
                cache_mode='overwrite'
            )
            elapsed = time.time() - start

            results.append({
                "region": region,
                "time_grouping": time_grouping,
                "elapsed": elapsed,
                "success": result is not None
            })

            print(
                f"Region: {region:15s} | Time grouping: {str(time_grouping):20s} | "
                f"Time: {elapsed:6.2f}s | Success: {result is not None}")

    print("\nPerformance Summary:")
    print("=" * 80)
    for r in results:
        print(f"{r['region']:15s} | {str(r['time_grouping']):20s} | {r['elapsed']:6.2f}s | {r['success']}")

    all_success = all(r["success"] for r in results)
    assert all_success, "Some metric computations failed"

    avg_time = sum(r["elapsed"] for r in results) / len(results)
    max_time = max(r["elapsed"] for r in results)
    min_time = min(r["elapsed"] for r in results)

    print(f"\nAverage time: {avg_time:.2f}s")
    print(f"Min time: {min_time:.2f}s")
    print(f"Max time: {max_time:.2f}s")


def test_grouped_metrics_performance_comprehensive():
    """Run a comprehensive performance test and print summary."""
    start_time = "2016-01-01"
    end_time = "2022-12-31"
    variable = "precip"
    agg_days = 7
    forecast = "imerg"
    truth = "ghcn_avg"
    # metric_name = "heidke-1-5-10-20"
    metric_name = "far-1"
    # metric_name = "mae"
    grid = "global1_5"
    mask = "lsm"
    region = "global"
    space_grouping = "country"
    # region = "countr"
    # time_grouping = "month"
    time_grouping = None

    results = []
    for i in range(1):
        start = time.time()
        result = metric(
            start_time=start_time,
            end_time=end_time,
            variable=variable,
            agg_days=agg_days,
            forecast=forecast,
            truth=truth,
            metric_name=metric_name,
            time_grouping=time_grouping,
            space_grouping=space_grouping,
            spatial=False,
            grid=grid,
            mask=mask,
            region=region,
            # recompute=['metric'],
            recompute=False,
            cache_mode='overwrite'
        )
        elapsed = time.time() - start

        plot_by_region(result, region, metric_name=metric_name)

        results.append({
            "region": region,
            "time_grouping": time_grouping,
            "elapsed": elapsed,
            "success": result is not None
        })

        print(
            f"Iteration {i}: Region: {region:15s} | Time grouping: {str(time_grouping):20s} | "
            f"Time: {elapsed:6.2f}s | Success: {result is not None}")

    print("\nPerformance Summary:")
    print("=" * 80)
    for r in results:
        print(f"{r['region']:15s} | {str(r['time_grouping']):20s} | {r['elapsed']:6.2f}s | {r['success']}")

    all_success = all(r["success"] for r in results)
    assert all_success, "Some metric computations failed"

    avg_time = sum(r["elapsed"] for r in results) / len(results)
    max_time = max(r["elapsed"] for r in results)
    min_time = min(r["elapsed"] for r in results)

    print(f"\nAverage time: {avg_time:.2f}s")
    print(f"Min time: {min_time:.2f}s")
    print(f"Max time: {max_time:.2f}s")


if __name__ == "__main__":
    # start_remote(remote_config=['xxxlarge_node', 'single_cluster'], remote_name='single_cluster_test')
    # test_grouped_metrics_performance_summary()
    # start_remote(remote_config='xlarge_cluster', remote_name='nuthatch_genevieve')
    start_remote(remote_config='xlarge_cluster')
    test_grouped_metrics_performance_comprehensive()
    # test_grouped_metrics_performance()



"""CHIRPS data product."""
import datetime

import fsspec
import pandas as pd
import xarray as xr
from dateutil import parser
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.utils import dask_remote, regrid, roll_and_agg
from sheerwater.interfaces import data as sheerwater_data, spatial


@dask_remote
@cache(cache_args=['year', 'grid', 'stations', 'version'],
       backend_kwargs={
           'chunking': {'lat': 300, 'lon': 300, 'time': 365}
})
def chirps_raw(year, grid, stations=True, version=2):  # noqa: ARG001
    """CHIRPS raw by year."""
    # Open the datastore
    if not stations and version == 3:
        base_url = f'https://data.chc.ucsb.edu/products/CHIRP-v3.0/daily/global/tifs/{year}/chirp-v3.0.'
        urls = []
        if year < 2000:
            print("Chirp v3 not valid before 2000. Returning none.")
            return None
        elif year == 2000:
            for date in pd.date_range(f"{year}-06-01", f"{year}-12-31"):
                date = date.date()
                full_url = base_url + f"{year}.{date.month:02}.{date.day:02}.tif"
                urls.append(full_url)
        else:
            fs = fsspec.filesystem("https")
            for date in pd.date_range(f"{year}-01-01", f"{year}-12-31"):
                date = date.date()
                full_url = base_url + f"{year}.{date.month:02}.{date.day:02}.tif"
                if fs.exists(full_url):
                    urls.append(full_url)

        def preprocess(ds):
            """Preprocess the dataset to add the member dimension."""
            ff = ds.encoding["source"]
            date = pd.to_datetime('-'.join(ff.split('/')[-1].split('.')[2:-1]))
            ds = ds.assign_coords(time=date)
            ds = ds.expand_dims(dim='time')
            ds = ds.squeeze('band')
            ds = ds.reset_coords('band', drop=True)
            return ds

        ds = xr.open_mfdataset(
            urls, engine='rasterio', preprocess=preprocess,
            chunks={'y': 1200, 'x': 1200, 'time': 365},
            concat_dim=["time"], compat="override", coords="minimal", combine="nested")
        ds = ds.rename({'y': 'lat', 'x': 'lon', 'band_data': 'precip'})
    elif stations and version == 2:
        if year == datetime.datetime.now().year:
            prelim_url = f'https://data.chc.ucsb.edu/products/CHIRPS-2.0/prelim/global_daily/netcdf/p05/chirps-v2.0.{year}.days_p05.nc'
            url = f'https://data.chc.ucsb.edu/products/CHIRPS-2.0/global_daily/netcdf/p05/chirps-v2.0.{year}.days_p05.nc'

            fs = fsspec.filesystem("https", timeout=7200)
            fobj = fs.open(url)
            fprelim = fs.open(prelim_url)
            ds = xr.open_dataset(fobj, chunks={'lat': 300, 'lon': 300, 'time': 365})
            ds_prelim = xr.open_dataset(fprelim, chunks={'lat': 300, 'lon': 300, 'time': 365})

            ds = ds.combine_first(ds_prelim)

            # Rename to lat/lon
            ds = ds.rename({'latitude': 'lat', 'longitude': 'lon'})
        else:
            url = f'https://data.chc.ucsb.edu/products/CHIRPS-2.0/global_daily/netcdf/p05/chirps-v2.0.{year}.days_p05.nc'

            fs = fsspec.filesystem("https", timeout=7200)
            fobj = fs.open(url)
            ds = xr.open_dataset(fobj, chunks={'lat': 300, 'lon': 300, 'time': 365})

            # Rename to lat/lon
            ds = ds.rename({'latitude': 'lat', 'longitude': 'lon'})
    elif stations and version == 3:
        base_url = f"https://data.chc.ucsb.edu/products/CHIRPS/v3.0/daily/final/sat/netcdf/byMonth/chirps-v3.0.{year}"
        urls = []
        if year < 2000:
            print("Chirps v3 not valid before 2000. Returning none.")
            return None
        elif year == 2000:
            for month in range(6, 13):
                full_url = base_url + f".{month:02}.days_p05.nc"
                urls.append(full_url)
        else:
            for month in range(1, 13):
                full_url = base_url + f".{month:02}.days_p05.nc"
                urls.append(full_url)

        fs = fsspec.filesystem("https", timeout=7200, block_size=10*10*1024*1024)
        files = []
        for url in urls:
            try:
                f = fs.open(url)
                files.append(f)
            except FileNotFoundError:
                pass

        ds = xr.open_mfdataset(files, chunks={'lat': 300, 'lon': 300, 'time': 365})

        # Rename to lat/lon
        ds = ds.rename({'latitude': 'lat', 'longitude': 'lon'})
    else:
        if not stations and version == 2:
            url = f'https://data.chc.ucsb.edu/products/CHIRP/daily/netcdf/chirp.{year}.days_p05.nc'
        else:
            raise ValueError(f"No chirps dataset found for stations:{stations}, version:{version}")

        fs = fsspec.filesystem("https", timeout=7200, block_size=10*10*1024*1024)
        fobj = fs.open(url, block_size=10*10*1024*1024)
        ds = xr.open_dataset(fobj, chunks={'lat': 300, 'lon': 300, 'time': 365})

        # Rename to lat/lon
        ds = ds.rename({'latitude': 'lat', 'longitude': 'lon'})

    # clip to the year
    start_time = str(year) + '-01-01'
    end_time = str(year) + '-12-31'

    ds = ds.sel({'time': slice(start_time, end_time)})

    return ds


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['grid', 'stations', 'version'],
       backend_kwargs={
           'chunking': {'lat': 300, 'lon': 300, 'time': 365}
})
def chirps_gridded(start_time, end_time, grid, stations=True, version=2,
                   mask=None, region='global'):  # noqa: ARG001
    """CHIRPS regridded by year."""
    years = range(parser.parse(start_time).year, parser.parse(end_time).year + 1)

    datasets = []
    for year in years:
        ds = chirps_raw(year, 'chirps', stations=stations, version=version, filepath_only=True)
        if ds is not None:
            datasets.append(ds)

    ds = xr.open_mfdataset(datasets,
                           engine='zarr',
                           parallel=True,
                           chunks={'lat': 300, 'lon': 300, 'time': 365})

    if "spatial_ref" in ds:
        ds = ds.drop_vars(["spatial_ref"])

    # Regrid if not on the native grid
    if grid != 'chirps':
        ds = regrid(ds, grid, base='base180', method='conservative', region=region)

    return ds


@timeseries()
@spatial()
@cache(cache_args=['grid', 'agg_days', 'stations', 'version'],
       cache_disable_if={'agg_days': 1},
       backend_kwargs={
           'chunking': {'lat': 300, 'lon': 300, 'time': 365}
})
def chirps_rolled(start_time, end_time, agg_days, grid, stations=True, version=2, mask=None, region='global'):
    """CHIRPS rolled and aggregated."""
    ds = chirps_gridded(start_time, end_time, grid, stations=stations, version=version, mask=mask, region=region)
    ds = roll_and_agg(ds, agg=agg_days, agg_col="time", agg_fn='mean')
    return ds


@dask_remote
def _chirps_unified(start_time, end_time, variable, agg_days, grid='global0_25',
                    stations=True, version=2, mask=None, region='global'):
    """A unified chirps caller."""
    if variable not in ['precip']:
        raise NotImplementedError("Only precip and derived variables provided by CHIRP/S.")
    ds = chirps_rolled(start_time, end_time, agg_days, grid,
                       stations=stations, version=version, mask=mask,
                       region=region)
    ds = ds.assign_attrs(sparse=True)
    return ds


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'])
def chirp_v2(start_time=None, end_time=None, variable='precip', agg_days=5,
             grid='global0_25', mask='lsm', region='global'):
    """A chirps interface for CHIRP2."""
    return _chirps_unified(start_time, end_time, variable, agg_days, grid=grid,
                           stations=False, version=2, mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False,
       cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'])
def chirp_v3(start_time=None, end_time=None, variable='precip', agg_days=5,
             grid='global0_25', mask='lsm', region='global'):
    """A chirps interface for CHIRP3."""
    return _chirps_unified(start_time, end_time, variable, agg_days, grid=grid,
                           stations=False, version=3, mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'])
def chirps_v2(start_time=None, end_time=None, variable='precip', agg_days=5,
              grid='global0_25', mask='lsm', region='global'):
    """A chirps interface for CHIRPS2."""
    return _chirps_unified(start_time, end_time, variable, agg_days, grid=grid,
                           stations=True, version=2, mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'])
def chirps_v3(start_time=None, end_time=None, variable='precip', agg_days=5,
              grid='global0_25', mask='lsm', region='global'):
    """A chirps interface for CHIRPS3."""
    return _chirps_unified(start_time, end_time, variable, agg_days, grid=grid,
                           stations=True, version=3, mask=mask, region=region)


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def chirps(start_time=None, end_time=None, variable='precip', agg_days=5,
           grid='global0_25', mask='lsm', region='global'):  # noqa: ARG001
    """Final access function for chirps."""
    ds = _chirps_unified(start_time, end_time, variable, agg_days, grid=grid,
                         stations=True, version=3, mask=mask, region=region)
    return ds



"""Get GHCND data."""
import math

import dask
import dask.dataframe as dd
import numpy as np
import pandas as pd
import xarray as xr
from dateutil import parser
from nuthatch import cache
from nuthatch.processors import timeseries

from sheerwater.utils import dask_remote, get_grid, get_grid_ds, get_variable, roll_and_agg, snap_point_to_grid
from sheerwater.interfaces import data as sheerwater_data, spatial


@cache(cache_args=[])
def ghcn_station_list():
    """Gets GHCN station metadata."""
    df = pd.read_table('https://www.ncei.noaa.gov/pub/data/ghcn/daily/ghcnd-inventory.txt',
                       sep="\\s+", names=['ghcn_id', 'lat', 'lon', 'unknown', 'start_year', 'end_year'])

    df = df.drop(['unknown'], axis=1)
    df = df.groupby(by=['ghcn_id']).first().reset_index()

    return df


@dask_remote
@cache(cache_args=['year', 'grid', 'cell_aggregation'],
       backend_kwargs={'chunking': {'time': 365, 'lat': 300, 'lon': 300}})
def ghcnd_yearly(year, grid='global0_25', cell_aggregation='first'):
    """Get a by year station data and save it as a zarr."""
    obs = dd.read_csv(f"s3://noaa-ghcn-pds/csv/by_year/{year}.csv",
                      names=['ghcn_id', 'date', 'variable', 'value', 'mflag', 'qflag', 'sflag', 'otime'],
                      header=0,
                      blocksize="32MB",
                      dtype={'ghcn_id': str,
                             'date': str,
                             'variable': str,
                             'value': int,
                             'mflag': str,
                             'qflag': str,
                             'sflag': str,
                             'otime': str},
                      storage_options={'anon': True},
                      on_bad_lines="skip")

    # Drop rows we don't care about
    obs = obs[obs['variable'].isin(['TMAX', 'TMIN', 'TAVG', 'PRCP'])]

    # Drop flagged data
    obs = obs[obs['qflag'].isna()]

    # Rplace any invalid data
    INVALID_NUMBER = 9999
    obs = obs.replace(INVALID_NUMBER, pd.NA)

    # Divide by 10 because data is represented in 10ths
    obs['value'] = obs['value'] / 10.0

    # Assign to new column based on variable values
    obs['tmax'] = obs.apply(lambda x: x.value if x['variable'] == 'TMAX' else pd.NA,
                            axis=1, meta=('tmax', 'f8'))
    obs['tmin'] = obs.apply(lambda x: x.value if x['variable'] == 'TMIN' else pd.NA,
                            axis=1, meta=('tmin', 'f8'))
    obs['temp'] = obs.apply(lambda x: x.value if x['variable'] == 'TAVG' else pd.NA,
                            axis=1, meta=('temp', 'f8'))
    obs['precip'] = obs.apply(lambda x: x.value if x['variable'] == 'PRCP' else pd.NA,
                              axis=1, meta=('precip', 'f8'))

    obs = obs.drop(['variable', 'value', 'qflag', 'mflag', 'sflag', 'otime'], axis=1)

    # Group by date and merge columns
    obs = obs.groupby(by=['date', 'ghcn_id']).first()
    obs = obs.reset_index()

    # If temp is none avrage the two
    atemp = (obs['tmin'] + obs['tmax'])/2
    obs['temp'] = obs['temp'].astype(float).fillna(atemp.astype(float))

    # Convert date into a datetime
    obs["time"] = dd.to_datetime(obs["date"])
    obs = obs.drop(['date'], axis=1)

    # Round the coordinates to the nearest grid
    lats, lons, grid_size, offset = get_grid(grid)

    stat = ghcn_station_list()
    stat['lat'] = stat['lat'].apply(lambda x: snap_point_to_grid(x, grid_size, offset))
    stat['lon'] = stat['lon'].apply(lambda x: snap_point_to_grid(x, grid_size, offset))

    stat = stat.set_index('ghcn_id')

    if 'start_year' in stat.columns:
        stat = stat.drop('start_year', axis=1)
    if 'end_year' in stat.columns:
        stat = stat.drop('end_year', axis=1)

    obs = obs.join(stat, on='ghcn_id', how='inner')

    if cell_aggregation == 'first':
        stations_to_use = obs.groupby(['lat', 'lon']).agg(ghcn_id=('ghcn_id', 'first'))
        stations_to_use = stations_to_use['ghcn_id'].unique()

        obs = obs[obs['ghcn_id'].isin(stations_to_use)]
        obs = obs.drop(['ghcn_id'], axis=1)
    elif cell_aggregation == 'mean':
        # Group by lat/lon/time
        obs = obs.groupby(by=['lat', 'lon', 'time']).agg(temp=('temp', 'mean'),
                                                         precip=('precip', 'mean'),
                                                         tmin=('tmin', 'min'),
                                                         tmax=('tmax', 'max'))
        obs = obs.reset_index()

    obs.temp = obs.temp.astype(np.float32)
    obs.tmax = obs.tmax.astype(np.float32)
    obs.tmin = obs.tmin.astype(np.float32)
    obs.precip = obs.precip.astype(np.float32)

    # Convert to xarray - for this to succeed obs must be a pandas dataframe
    obs = xr.Dataset.from_dataframe(obs.compute().set_index(['time', 'lat', 'lon']))

    # Reindex to fill out the lat/lon
    grid_ds = get_grid_ds(grid)
    obs = obs.reindex_like(grid_ds, method='nearest', tolerance=0.005)

    # Return the xarray
    return obs


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['grid', 'cell_aggregation'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def ghcnd(start_time, end_time, grid="global0_25", cell_aggregation='first',
          mask=None, region='global', delayed=True):  # noqa: ARG001
    """Final gridded station data before aggregation."""
    # Get years between start time and end time
    years = range(parser.parse(start_time).year, parser.parse(end_time).year + 1)

    datasets = []
    for year in years:
        if delayed:
            ds = dask.delayed(ghcnd_yearly)(year, grid, cell_aggregation, filepath_only=True)
            datasets.append(ds)
        else:
            ds = dask.delayed(ghcnd_yearly)(year, grid, cell_aggregation, filepath_only=True)
            datasets.append(dask.compute(ds)[0])

    if delayed:
        datasets = dask.compute(*datasets)

    x = xr.open_mfdataset(datasets,
                          engine='zarr',
                          parallel=True,
                          chunks={'lat': 300, 'lon': 300, 'time': 365})

    return x


@dask_remote
@timeseries()
@spatial()
@cache(cache_args=['grid', 'agg_days', 'missing_thresh', 'cell_aggregation'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def ghcnd_rolled(start_time, end_time, agg_days,
                 grid='global0_25', missing_thresh=0.9, cell_aggregation='first', mask=None, region='global'):
    """GHCND rolled and aggregated."""
    # Get the data
    ds = ghcnd(start_time, end_time, grid, cell_aggregation, mask=mask, region=region)

    # Roll and agg
    agg_thresh = max(math.ceil(agg_days*missing_thresh), 1)
    ds = roll_and_agg(ds, agg=agg_days, agg_col="time", agg_fn='mean', agg_thresh=agg_thresh)
    return ds


@dask_remote
def _ghcn_unified(start_time, end_time, variable, agg_days,
                  grid='global0_25', missing_thresh=0.9, cell_aggregation='first', mask=None, region='global'):
    """Standard interface for ghcn data."""
    ds = ghcnd_rolled(start_time, end_time, agg_days=agg_days, grid=grid,
                      missing_thresh=missing_thresh, cell_aggregation=cell_aggregation, mask=mask, region=region)

    # Get the variable
    variable_ghcn = get_variable(variable, 'ghcn')
    variable_sheerwater = get_variable(variable, 'sheerwater')
    ds = ds[variable_ghcn].to_dataset()
    # Rename
    ds = ds.rename({variable_ghcn: variable_sheerwater})
    # Note that this is sparse
    ds = ds.assign_attrs(sparse=True)
    return ds


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region', 'missing_thresh'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def ghcn(start_time=None, end_time=None, variable='precip', agg_days=1,
         grid='global0_25', mask='lsm', region='global', missing_thresh=0.9):
    """Standard interface for ghcn data."""
    return _ghcn_unified(start_time, end_time, variable, agg_days=agg_days,
                         grid=grid, mask=mask, region=region,
                         missing_thresh=missing_thresh, cell_aggregation='first')


@dask_remote
@sheerwater_data()
@cache(cache=False, cache_args=['variable', 'agg_days', 'grid', 'mask', 'region', 'missing_thresh'],
       backend_kwargs={'chunking': {'lat': 300, 'lon': 300, 'time': 365}})
def ghcn_avg(start_time=None, end_time=None, variable='precip', agg_days=1,
             grid='global0_25', mask='lsm', region='global', missing_thresh=0.9):
    """Standard interface for ghcn data."""
    return _ghcn_unified(start_time, end_time, variable, agg_days=agg_days,
                         grid=grid, mask=mask, region=region,
                         missing_thresh=missing_thresh, cell_aggregation='mean')
